from collections.abc import Sequence
from dataclasses import dataclass

from inline_snapshot import snapshot

from memium.destination.ankiconnect.anki_formatter import AnkiQAFormatter
from memium.destination.ankiconnect.anki_model import AnkiQAModel


def fake_tag_factory() -> Sequence[str]:
    return ["FakeTag"]


import pytest


@dataclass(frozen=True)
class QAExample:
    card: AnkiQAModel
    expected_deck: str


@pytest.mark.parametrize(
    ("example"),
    [
        QAExample(AnkiQAModel.dummy(tags=["anki/deck/Subdeck"]), "FakeBaseDeck::Subdeck"),
        QAExample(
            AnkiQAModel.dummy(
                tags=["anki/deck/Subdeck"], question="What are _Wikilinks_ on _Wikipedia_?"
            ),
            "FakeBaseDeck::Subdeck::Wikilinks-Wikipedia",
        ),
        QAExample(
            AnkiQAModel.dummy(question="On _Wikipedia_, what are _Wikilinks_?"),
            "FakeBaseDeck::Wikilinks-Wikipedia",
        ),
        QAExample(AnkiQAModel.dummy(question="Without wikilinks?"), "FakeBaseDeck"),
    ],
    ids=lambda x: x.expected_deck,
)
def test_ankiqa_deck_inference(example: QAExample):
    assert example.card.deck == example.expected_deck


def test_model():
    card = AnkiQAModel.dummy(question="Q. This is a _question_?")
    model = AnkiQAFormatter("FakeCSS").format(card).model  # type: ignore
    assert model.model_id == snapshot(9079488902)  # type: ignore
    assert model.model_type == snapshot(0)  # type: ignore
    assert model.name == snapshot('Ankdown QA with raw text')  # type: ignore
    assert model.fields == snapshot(  # type: ignore
        [{"name": "Question"}, {"name": "Answer"}, {"name": "Extra"}, {'name':'raw_question'}, {'name':'raw_answer'}]
    )  # type: ignore
    assert model.css == snapshot('FakeCSS')  # type: ignore
    assert model.templates == snapshot(  # type: ignore
        [
            {
                "name": 'Ankdown QA with raw text',
                "qfmt": """\

<div class="front">{{ Extra }}
    {{ Question }}
</div>\
""",
                "afmt": """\

<div class="back">{{ Extra }}
    <div class="question">
        {{ Question }}
    </div>
    <div class="answer">
        {{ Answer }}
    </div>
</div>
            \
""",
            }
        ]
    )  # type: ignore


def test_formatting():
    card = AnkiQAModel.dummy(question="Q. This is a _question_?")
    note = AnkiQAFormatter("").format(card)
    assert note.fields == snapshot(  # type: ignore
        [
            "<p>Q. This is a <em>question</em>?</p>",
            '<p>A. DummyAnswer</p>',
            '', 'Q. This is a _question_?', 'A. DummyAnswer']
    )



from collections.abc import Mapping, Sequence
from pathlib import Path

import genanki
import pytest

from ...environment import get_host_home_dir
from .ankiconnect_gateway import (
    ANKICONNECT_URL,
    AnkiConnectGateway,
    AnkiField,
    NoteInfo,
    anki_connect_is_live,
)


class MockNoteInfo(NoteInfo):
    noteId: int = 1
    tags: Sequence[str] = ["MockTag"]
    fields: Mapping[str, AnkiField] = {
        "Text": AnkiField(value="MockText", order=0),
        "raw_question": AnkiField(value="MockQuestion", order=1),
        "raw_answer": AnkiField(value="MockAnswer", order=2),
    }
    modelName: str = "MockModel"
    cards: Sequence[int] = [1]


@pytest.mark.skipif(
    not anki_connect_is_live() or not Path("/output").exists(),
    reason="Tests require a running AnkiConnect server and an output directory. Use the Docker container to run the tests.",
)
class TestAnkiConnectGateway:
    output_path = Path("/output")

    def test_import_get_delete_happy_path(self):
        # Delete all .apkg in the output directory
        for f in self.output_path.glob("*.apkg"):
            f.unlink()

        deck = genanki.Deck(deck_id=1, name="Test deck")
        model = genanki.Model(
            model_id=1,
            name="Test model",
            fields=[{"name": "Question"}, {"name": "Answer"}],
            templates=[
                {
                    "name": "Card 1",
                    "qfmt": "{{Question}}",
                    "afmt": "{{FrontSide}}<hr id=answer>{{Answer}}",
                }
            ],
        )

        deck.add_model(model)  # type: ignore
        deck.add_note(  # type: ignore
            genanki.Note(
                model=model, fields=["Capital of Argentina", "Buenos Aires"], tags=["TestTag"]
            )
        )

        tmp_read_dir = get_host_home_dir() / "ankidecks"
        gateway = AnkiConnectGateway(
            ankiconnect_url=ANKICONNECT_URL,
            root_deck="Test deck",
            tmp_read_dir=tmp_read_dir,
            tmp_write_dir=self.output_path,
            max_deletions_per_run=1,
            max_wait_seconds=0,
        )

        # Phase 1: Importing
        package = genanki.Package(deck_or_decks=deck)
        gateway.import_package(package=package)
        assert self.output_path.is_dir()
        assert len(list(Path("/output").glob("*.apkg"))) == 0

        # Phase 2: Getting
        all_notes = gateway.get_all_note_infos()
        note = all_notes[0]
        assert note.tags == ["TestTag"]

        # Phase 3: Deleting
        gateway.delete_notes(note_ids=[n.noteId for n in all_notes])
        assert len(gateway.get_all_note_infos()) == 0

    def test_error_if_deleting_more_than_allowed(self):
        gateway = AnkiConnectGateway(
            ankiconnect_url=ANKICONNECT_URL,
            root_deck="Test deck",
            tmp_read_dir=Path("/tmp"),
            tmp_write_dir=Path("/tmp"),
            max_deletions_per_run=0,
            max_wait_seconds=0,
        )
        with pytest.raises(ValueError, match="are scheduled for deletion"):
            gateway.delete_notes(note_ids=[1])


def test_error_if_not_running():
    with pytest.raises(ConnectionError, match="Could not connect to Anki"):
        AnkiConnectGateway(
            ankiconnect_url="http://localhost:1234",
            root_deck="Test deck",
            tmp_read_dir=Path("/tmp"),
            tmp_write_dir=Path("/tmp"),
            max_deletions_per_run=0,
            max_wait_seconds=0,
        )



import pathlib
from collections.abc import Sequence
from dataclasses import dataclass

import pytest
from inline_snapshot import snapshot

from memium.utils.markdown import get_terms_surrounded_by_underscores

from .markdown import md_to_html


@dataclass(frozen=True)
class Ex:
    given: str
    # Required setup

    then: str
    # What the expected result is


@pytest.mark.parametrize(
    ("example"),
    [
        Ex("_Method_s", "<em>Method</em>s")  # Example one
    ],
)
def test_markdown_parser(example: Ex):
    assert md_to_html(example.given) == f"<p>{example.then}</p>"


def test_code_block_parsing(tmp_path: pathlib.Path):
    contents = pathlib.Path(__file__).parent / "code_block.md"
    parsed = md_to_html(contents.read_text())
    path = tmp_path / "test.htm"
    path.write_text(parsed)
    assert parsed == snapshot("""\
<pre><code class="language-python">def myfunc():
    return res
</code></pre>\
""")


@dataclass(frozen=True)
class Example:
    input_string: str
    wikilinks: Sequence[str]


@pytest.mark.parametrize(
    ("example"),
    [
        Example("_Test_", ["Test"]),
        Example("_Test (testing)_", ["Test (testing)"]),
        Example("_Test_, _Test2_", ["Test", "Test2"]),
        Example("_ Some things, are just not _Wikilinks_ _", ["Wikilinks"]),
    ],
    ids=lambda x: x.input_string,
)
def test_get_wikilinks_from_string(example: Example):
    assert get_terms_surrounded_by_underscores(example.input_string) == example.wikilinks



from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import quote

from memium.source.document import Document

from ..utils.hash_cleaned_str import clean_str, hash_str_to_int


def obsidian_url(file_title: str, line_nr: int | None) -> str:
    url = f"obsidian://advanced-uri?filename={quote(file_title)}"
    if line_nr:
        url += f"&line={line_nr}"
    return url


@dataclass(frozen=True)
class QAPrompt:
    """Value object for a question-answer prompt. Its identity is determined entirely by its values."""

    question: str
    answer: str

    def __post_init__(self):
        """Styling should not occur in this value object, since any change to style would change its scheduling identity.

        Instead, styling should happen in the destination (e.g. AnkiQA)."""
        forbidden = ["style="]

        errors: list[str] = []
        for tag in forbidden:
            if tag in self.question:
                errors.append(f"'{tag}' found in question")
            if tag in self.answer:
                errors.append(f"'{tag}' found in answer")

        if errors:
            error_str = "\n".join(errors)
            raise ValueError(error_str)

    @staticmethod
    def dummy(question: str = "DummyQuestion", answer: str = "DummyAnswer") -> "QAPrompt":
        return QAPrompt(question=question, answer=answer)

    @property
    def scheduling_uid_str(self) -> str:
        """Str used for generating the update_uid. Super helpful for debugging."""
        return cleaned_qa_scheduling_uid_str(self.question, self.answer)

    @property
    def scheduling_uid(self) -> int:
        return hash_str_to_int(self.scheduling_uid_str)


@dataclass(frozen=True)
class QAWithDoc:
    prompt: QAPrompt

    parent_doc: Document

    line_nr: int

    render_parent_doc: bool = True

    @staticmethod
    def dummy(
        question: str = "FakeQuestion",
        answer: str = "FakeAnswer",
        parent_doc: Document | None = None,
        line_nr: int = 0,
    ) -> "QAWithDoc":
        return QAWithDoc(
            prompt=QAPrompt.dummy(question, answer),
            parent_doc=Document("DummyContent", Path("DummyPath"))
            if parent_doc is None
            else parent_doc,
            line_nr=line_nr,
        )

    @property
    def scheduling_str(self) -> str:
        return self.prompt.scheduling_uid_str

    @property
    def scheduling_id(self) -> int:
        return self.prompt.scheduling_uid

    @property
    def tags(self) -> Sequence[str]:
        return self.parent_doc.tags

    @property
    def edit_url(self) -> str:
        return obsidian_url(self.parent_doc.title, self.line_nr)


def cleaned_qa_scheduling_uid_str(question: str, answer: str) -> str:
    return f"{clean_str(question)}_{clean_str(answer)}"


def qa_update_uid_str(question: str, answer: str, tags: Sequence[str]) -> str:
    return f"{question}_{answer}_{tags}"



from collections.abc import Sequence
from dataclasses import dataclass
from typing import NewType

from memium.source.prompt import QAPrompt
from memium.utils.markdown import get_terms_surrounded_by_underscores

Markdown = NewType("Markdown", str)


@dataclass(frozen=True)
class AnkiQAModel:
    # Field capitalisation must be preserved for backwards compatibility
    Question: Markdown
    Answer: Markdown
    Extra: Markdown

    raw_prompt: QAPrompt

    tags: Sequence[str]
    root_deck: str

    @staticmethod
    def dummy(
        question: str = "Q. DummyQuestion?",
        answer: str = "A. DummyAnswer",
        extra: str = "",
        tags: Sequence[str] = (),
    ) -> "AnkiQAModel":
        return AnkiQAModel(
            Question=Markdown(question),
            Answer=Markdown(answer),
            Extra=Markdown(extra),
            tags=tags,
            raw_prompt=QAPrompt.dummy(question, answer),
            root_deck="FakeBaseDeck",
        )

    @property
    def formatted_field_names(self) -> list[str]:
        return [name for name, type_hint in self.__annotations__.items() if type_hint == Markdown]

    @property
    def deck(self) -> str:
        deck_prefix = "anki/deck/"
        deck_in_tags = (
            tag.replace(deck_prefix, "").replace("/", "::")
            for tag in self.tags
            if tag.startswith(deck_prefix)
        )
        subdeck = next(deck_in_tags, None)

        base_anki_deck = self.root_deck if subdeck is None else f"{self.root_deck}::{subdeck}"

        wiki_links = get_terms_surrounded_by_underscores(self.Question)
        wiki_subdeck = "-".join(sorted(wiki_links))

        if wiki_subdeck != "":
            return f"{base_anki_deck}::{wiki_subdeck}"
        return base_anki_deck



import re
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class Document:
    content: str
    source_path: Path

    @staticmethod
    def dummy(
        content: str | None = None, source_path: Path = Path("DummyPath"), tags: Sequence[str] = ()
    ) -> "Document":
        return Document(
            content=content
            if content is not None
            else "dummy content with tags: " + ", ".join(f"#{tag}" for tag in tags),
            source_path=source_path,
        )

    @property
    def tags(self) -> Sequence[str]:
        tag_strings: list[str] = list(re.findall(r"#[\w\/]+", self.content))
        return [tag_string.replace("#", "") for tag_string in tag_strings]

    @property
    def title(self) -> str:
        return self.source_path.stem

    def __repr__(self) -> str:
        return f"{self.title}: {self.content[0:10]}..."


@dataclass(frozen=True)
class DummyDocument(Document):
    content: str = "dummy content"
    source_path: Path = Path("dummy_path")



from collections.abc import Sequence
from pathlib import Path

import invoke as inv


def get_code_blocks_from_md(md_path: Path) -> Sequence[str]:
    md = md_path.read_text()
    code_blocks = md.split("```")[1::2]
    return code_blocks


def create_smoketest_dir() -> Path:
    print("💨 Running smoketest")
    input_dir = Path() / ".smoketest_dir"
    input_dir.mkdir(exist_ok=True)

    test_file = input_dir / "test.md"
    test_file.write_text("Q. Question here\nA. Answer!")
    return input_dir


@inv.task  # type: ignore
def smoketest_cli(c: inv.Context):
    smoketest_dir = create_smoketest_dir()
    code_blocks = get_code_blocks_from_md(Path("readme.md"))
    cli_block = next(block for block in code_blocks if "cli-block" in block)
    sanitised_cli_block = cli_block.splitlines()[1].replace("> ", "")

    cli_smoketest_cmd = sanitised_cli_block.replace("[YOUR_INPUT_DIR]", str(smoketest_dir))
    print(cli_smoketest_cmd)
    c.run(cli_smoketest_cmd + "  --dry-run \\\n" + "  --skip-sync")
    print("💨🎉 Smoketest complete")



from collections.abc import Sequence
from typing import Protocol

from memium.source.prompt import QAWithDoc

from ..document import Document


class BasePromptExtractor(Protocol):
    def extract_prompts(self, document: Document) -> Sequence[QAWithDoc]: ...
